// **************************************************************************************
// el Comandante                                                    guerilla.bplaced.net
// --------------------------------------------------------------------------------------
// Information:
// Copies ALL files from the specified SD-Card directory to SPIFFS
// Long filenames will be truncated to 30 characters (SPIFFS filename limitation)
// If a file(name) already esist in SPIFFS it will be overwritten
// Take care that there is enough free space in SPIFFS for the files to copy
// --------------------------------------------------------------------------------------
// Used Libraries:
// TFT_eSPI --> https://github.com/Bodmer/TFT_eSPI
// **************************************************************************************

#include <SPIFFS.h>                                       // SPIFFS filesystem library
#include <SPI.h>                                          // SPI interface library
#include <SD.h>                                           // SD-Card library
#include <TFT_eSPI.h>                                     // TFT_espi library
TFT_eSPI tft = TFT_eSPI();

// ######################################################################################
// Change values in this area to match your envoirement and desires
// --------------------------------------------------------------------------------------
// Custom VSPI Pins for the SD-Card breakout board - must match the hardware
#define   SCK   27
#define   MISO  32
#define   MOSI  33
#define   CS    2
// Source directory on the SD-Card for the files copied to SPIFFS
String    sdcardDirectory = "/Temp2762";
// Format SPIFFS before copy, takes ~12 seconds, 0 = don't format, 1 = format
uint8_t   spiffsFormat = 0;
// Delay in ms for messages, even if insufficient storage for a file in SPIFFS
uint16_t  messageDelay = 3000;
// ######################################################################################

// Vars
uint8_t   buffer[4096];
uint32_t  update;
uint64_t  TotalBytesToCopy;
uint32_t  i;
uint32_t  z;
uint32_t  tftWidth = 0;
uint32_t  tftHeight = 0;
uint32_t  tftBaseY = 8;
uint32_t  tftSpaceY = 4;
uint32_t  fontHeight = 0;
// Screen elements position     = SPIFFS, SD-Card, H-Line, SPI Status, SD-Card Type, SD-Card Size, SD-Card Free, SD-Card nFiles, SPIFFS Status, SPIFFS Type, SPIFFS Size, SPIFFS Free, SPIFFS nFiles, H-Line, Message, Progress
uint32_t  ElementsArrayXpos[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
uint32_t  ElementsArrayYpos[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
String    szTemp = "";

// Vars - VSPI SD-Card
uint32_t  vspiBusFrequency = 2000000;                     // >= 4000000 sd-card mount could fail
File      sdRoot;                                         // SD root variable
File      sdFile;                                         // SD file variable
uint8_t   spiStatus;                                      // SPI status
char      spiStatusArray[2][13] = {"Mount: Error", "Mount: Okay"};
uint8_t   sdcardType;
char      sdcardArrayType[6][14] = {"Type: None", "Type: MMC", "Type: SDSC", "Type: SDHC", "Type: Unknown"};
uint64_t  sdcardSize;
uint64_t  sdcardDirSize;
uint64_t  sdcardFree;
uint32_t  sdcardFiles;

// Vars  - SPIFFS
File      spiffsRoot;                                     // SPIFFS root variable
File      spiffsFile;                                     // SPIFFS file variable
uint8_t   spiffsStatus;                                   // SPIFFS status
char      spiffsStatusArray[2][14] = {"Status: Error", "Status: Okay"};
uint8_t   spiffsType;
char      spiffsArrayType[6][14] = {"Type: SPIFFS", "Type: 1", "Type: 2", "Type: 3", "Type: Unknown"};
uint64_t  spiffsSize;
uint64_t  spiffsUsed;
uint64_t  spiffsFree;
uint32_t  spiffsFiles;

// Create a new SPI class on HSPI or VSPI 
SPIClass vspi = SPIClass(VSPI);

// Setup
void setup()
{
  pinMode(CS, OUTPUT);                                    // set up slave select pin CS as output
  SpiffsInit();                                           // init SPIFFS
  VspiInit();                                             // init VSPI before TFT_eSPI
  tft.init();                                             // init TFT_eSPI
  digitalWrite(TFT_BL, 0);                                // switch backlight off
  tft.setRotation(3);                                     // set the desired rotation
  ScreenSize();

  if (spiffsStatus == 1)
  {
    SpiffsFormat();
    SpiffsSizeSpaceFree();
    SpiffsFiles();
  }

  if (spiStatus == 1)
  {
    SDCardType();
    SDCardSize();
    SDCardDirSize();
  }

  ScreenBuild();
  tft.setTextDatum(TC_DATUM);
  digitalWrite(TFT_BL, 1);                                // switch backlight on

  if (spiffsStatus == 0 || spiStatus == 0 || sdcardType == 0)
  {
    tft.drawString("SPIFFS Error or SD-Card Error...", ElementsArrayXpos[14], ElementsArrayYpos[14]);
  }
  else
  {
    if (spiffsFree < sdcardDirSize)
    {
      szTemp = "Storage Low - Transfer in  ";
      tft.drawString(szTemp, ElementsArrayXpos[14], ElementsArrayYpos[14]);
      i = 9;
    }
    else
    {
      szTemp = "Storage Okay - Transfer in  ";
      tft.drawString(szTemp, ElementsArrayXpos[14], ElementsArrayYpos[14]);
      i = 5;
    }
    while (i > 0)
    {
      if (update <= millis())
      {
        update = millis() + 1000;        
        szTemp.setCharAt(szTemp.length() - 1, i + '0');
        tft.drawString(szTemp, ElementsArrayXpos[14], ElementsArrayYpos[14]);
        i--;
      }
    }
    CopySDCardFilesToSPIFFS();
    delay(messageDelay);
    tft.fillRect(0, ElementsArrayYpos[14], tftWidth, fontHeight, 0x0000);
    tft.setCursor(ElementsArrayXpos[14] / 1.35, ElementsArrayYpos[14]);
    tft.printf("%d / %d files copied", i, sdcardFiles);
    delay(messageDelay);
    ListSPIFFS();
  }
}


// Loop
void loop()
{
  // notting to do
}


// **************************************************************************************
// Screen - size
void ScreenSize()
{
  tftWidth = tft.width();                                 // get tft width
  tftHeight = tft.height();                               // get tft height
  if (tftWidth >= 320 || tftHeight >= 320)                // big screen (i.e. WT32-SC01 - 480 x 320)
  {
    tft.setTextFont(2);
    fontHeight = tft.fontHeight(2);                       // get the font height
    ElementsArray();
  }
  else if (tftWidth < 320 || tftHeight < 320)             // small screen (i.e. LILYGO TTGO-T-Display - 240 x 135)
  {
    tft.setTextFont(1);
    fontHeight = tft.fontHeight(1);                       // get the font height
    ElementsArray();    
  }
}

// **************************************************************************************
// Screen - fill elements array with Xpos / Ypos
void ElementsArray()
{
  ElementsArrayXpos[0]  = 0;                              // SD-Card text
  ElementsArrayYpos[0]  = tftBaseY;
  ElementsArrayXpos[1]  = tftWidth / 2;                   // SPIFFS text
  ElementsArrayYpos[1]  = tftBaseY;

  ElementsArrayXpos[2]  = 0;                              // horizontal line
  ElementsArrayYpos[2]  = tftBaseY + fontHeight * 1 + tftSpaceY * 1;

  ElementsArrayXpos[3]  = 0;                              // SPI status
  ElementsArrayYpos[3]  = tftBaseY + fontHeight * 2 + tftSpaceY * 1;
  ElementsArrayXpos[4]  = 0;                              // SD-Card type
  ElementsArrayYpos[4]  = tftBaseY + fontHeight * 3 + tftSpaceY * 2;
  ElementsArrayXpos[5]  = 0;                              // SD-Card size
  ElementsArrayYpos[5]  = tftBaseY + fontHeight * 4 + tftSpaceY * 3;
  ElementsArrayXpos[6]  = 0;                              // SD-Card free
  ElementsArrayYpos[6]  = tftBaseY + fontHeight * 5 + tftSpaceY * 4;
  ElementsArrayXpos[7]  = 0;                              // SD-Card nfiles
  ElementsArrayYpos[7]  = tftBaseY + fontHeight * 6 + tftSpaceY * 5;

  ElementsArrayXpos[8]  = tftWidth / 2;                   // SPIFFS status
  ElementsArrayYpos[8]  = tftBaseY + fontHeight * 2 + tftSpaceY;
  ElementsArrayXpos[9]  = tftWidth / 2;                   // SPIFFS type
  ElementsArrayYpos[9]  = tftBaseY + fontHeight * 3 + tftSpaceY * 2;
  ElementsArrayXpos[10] = tftWidth / 2;                   // SPIFFS size
  ElementsArrayYpos[10] = tftBaseY + fontHeight * 4 + tftSpaceY * 3;
  ElementsArrayXpos[11] = tftWidth / 2;                   // SPIFFS free
  ElementsArrayYpos[11] = tftBaseY + fontHeight * 5 + tftSpaceY * 4;
  ElementsArrayXpos[12] = tftWidth / 2;                   // SPIFFS nfiles
  ElementsArrayYpos[12] = tftBaseY + fontHeight * 6 + tftSpaceY * 5;

  ElementsArrayXpos[13] = 0;                              // horizontal line
  ElementsArrayYpos[13] = tftBaseY + fontHeight * 9 + tftSpaceY * 2;

  ElementsArrayXpos[14] = tftWidth / 2;                   // message text
  ElementsArrayYpos[14] = tftBaseY + fontHeight * 9 + tftSpaceY * 3.5;

  ElementsArrayXpos[15] = 0;                              // progress bar
  ElementsArrayYpos[15] = tftBaseY + fontHeight * 12 + tftSpaceY / 2;
}


// **************************************************************************************
// Screen - build
void ScreenBuild()
{
  tft.fillScreen(0x0000);
  tft.setTextColor(0x07FF, 0x0000);
  tft.setTextSize(1);
  tft.setTextDatum(TL_DATUM);
  tft.drawString("SD-Card", ElementsArrayXpos[0], ElementsArrayYpos[0]);
  tft.drawString("SPIFFS", ElementsArrayXpos[1], ElementsArrayYpos[1]);

  tft.drawLine(ElementsArrayXpos[2], ElementsArrayYpos[2], tftWidth, ElementsArrayYpos[2], 0x8410);

  tft.setTextColor(0xFFFE, 0x0000);
  tft.drawString(spiStatusArray[spiStatus], ElementsArrayXpos[3], ElementsArrayYpos[3]);
  tft.drawString(sdcardArrayType[sdcardType], ElementsArrayXpos[4], ElementsArrayYpos[4]);
  tft.setCursor(ElementsArrayXpos[5], ElementsArrayYpos[5]);
  tft.printf("Size: %lld MB", sdcardSize / (1024 * 1024));
  tft.setCursor(ElementsArrayXpos[6], ElementsArrayYpos[6]);
  if (sdcardDirSize >= 1024)
  {tft.printf("Path: %lld KB", sdcardDirSize / 1024);}
  else
  {tft.printf("Path: %lld Byte", sdcardDirSize);}
  tft.setCursor(ElementsArrayXpos[7], ElementsArrayYpos[7]);
  tft.printf("Files: %d  ", sdcardFiles);

  tft.drawString(spiffsStatusArray[spiffsStatus], ElementsArrayXpos[8], ElementsArrayYpos[8]);
  tft.drawString(spiffsArrayType[spiffsType], ElementsArrayXpos[9], ElementsArrayYpos[9]);
  tft.setCursor(ElementsArrayXpos[10], ElementsArrayYpos[10]);
  tft.printf("Size: %lld KB", spiffsSize / 1024);
  tft.setCursor(ElementsArrayXpos[11], ElementsArrayYpos[11]);
  tft.printf("Free: %lld KB", spiffsFree / 1024);
  tft.setCursor(ElementsArrayXpos[12], ElementsArrayYpos[12]);
  tft.printf("Files: %d  ", spiffsFiles);

  tft.drawLine(ElementsArrayXpos[13], ElementsArrayYpos[13], tftWidth, ElementsArrayYpos[13], 0x8410);
}


// **************************************************************************************
// SPIFFS - init 
void SpiffsInit()
{
  if (!SPIFFS.begin(true))
  {spiffsStatus = 0;}
  else
  {spiffsStatus = 1;}
}

// **************************************************************************************
// SPIFFS - format
void SpiffsFormat()
{
  if (spiffsFormat == 1)
  {
    tft.fillScreen(0x0000);
    digitalWrite(TFT_BL, 1);                              // switch backlight on
    tft.setTextDatum(TC_DATUM);
    tft.drawString("Formating SPIFFS...", ElementsArrayXpos[14], ElementsArrayYpos[14]);
    while(!SPIFFS.format())                               // wait until format completed
    {};
    digitalWrite(TFT_BL, 0);                              // switch backlight off
  }
}

// **************************************************************************************
// SPIFFS - get size
void SpiffsSizeSpaceFree()
{
  spiffsSize = SPIFFS.totalBytes();
  spiffsUsed = SPIFFS.usedBytes();
  spiffsFree = SPIFFS.totalBytes() - SPIFFS.usedBytes();
}


// **************************************************************************************
void SpiffsFiles()
{
  spiffsRoot = SPIFFS.open("/");
  while(spiffsRoot.openNextFile())
  {spiffsFiles++;}
  spiffsRoot.close(); 
}


// **************************************************************************************
// SPI - Init communication protocol and the SD-Card before TFT_eSPI, so the chip select is defined
void VspiInit()
{
  vspi.begin(SCK, MISO, MOSI, CS);
  if (!SD.begin(CS, vspi, vspiBusFrequency))
  {spiStatus = 0;}
  else
  {spiStatus = 1;}
}


// **************************************************************************************
// SD-Card - get type
void SDCardType()
{
  if(spiStatus == false)                                  // SPI - check status
  {return;}
  z = SD.cardType();
  if(z == CARD_NONE)
  {sdcardType = 0;}
  else if(z == CARD_MMC)
  {sdcardType = 1;}
  else if(z == CARD_SD)
  {sdcardType = 2;}
  else if(z == CARD_SDHC)
  {sdcardType = 3;}
  else
  {sdcardType = 4;}
}

// **************************************************************************************
// SD-Card - get size and free space
void SDCardSize()
{sdcardSize = SD.cardSize();}                             // card size in bytes


// **************************************************************************************
// SD-Card - get free space
void SDCardDirSize()
{
  sdRoot = SD.open(sdcardDirectory);
  while (true)
  {
    sdFile = sdRoot.openNextFile();
    if (!sdFile)                                          // no more entry
    {break;}
    if (!sdFile.isDirectory())                            // skip directories
    {
      sdcardFiles++;
      sdcardDirSize = sdcardDirSize + sdFile.size();
    }
    sdFile.close();
  }
  sdRoot.close();
}


// **************************************************************************************
void CopySDCardFilesToSPIFFS()
{
  tft.setTextDatum(TC_DATUM);  
  if (sdcardFiles > 0)
  {
    i = 0;
    z = 0;
    sdRoot = SD.open(sdcardDirectory);
    sdFile = sdRoot.openNextFile();
    while (sdFile)
    {
      if (!sdFile.isDirectory())                          // skip directories
      {
        szTemp = sdFile.name();
        tft.fillRect(0, ElementsArrayYpos[14], tftWidth, fontHeight, 0x0000);
        tft.drawString(szTemp, ElementsArrayXpos[14], ElementsArrayYpos[14]);
        if (SPIFFS.totalBytes() - SPIFFS.usedBytes() >= sdFile.size())
        {
          tft.fillRect(ElementsArrayXpos[15] + tftWidth / sdcardFiles * z, ElementsArrayYpos[15], tftWidth / sdcardFiles, 8, 0x07FF);
          if (szTemp.length() >= 30)                      // SPIFFS filename limitation
          {
            szTemp = szTemp.substring(szTemp.length() - 30);
            szTemp.setCharAt(0, '~');
          }
          spiffsFile = SPIFFS.open("/" + szTemp, "w");
          while(sdFile.available())
          {spiffsFile.write(buffer, sdFile.read(buffer, 4096));}
          spiffsFile.close();
          i++;
        }
        else
        {
          tft.fillRect(ElementsArrayXpos[15] + tftWidth / sdcardFiles * z, ElementsArrayYpos[15], tftWidth / sdcardFiles, 8, 0xF800);
          delay(messageDelay);
        }
      }
      z++;
      sdFile = sdRoot.openNextFile();
    }
    sdRoot.close();
  }
  else
  {
    tft.fillRect(0, ElementsArrayYpos[14], tftWidth, fontHeight, 0x0000);
    tft.drawString("No Files in SD-Card Directory", ElementsArrayXpos[14], ElementsArrayYpos[14]);
  }
}


// **************************************************************************************
// SPIFFS - list files
void ListSPIFFS()
{
  tft.setRotation(0);
  tft.fillScreen(0x0000);
  tft.setTextColor(0xFFFE, 0x0000);
  tft.setCursor(0, 0);
  tft.setTextDatum(TL_DATUM);
  tft.println(" Files in SPIFFS");
  tft.println(" ------------------");
  tft.setTextFont(1);
  spiffsRoot = SPIFFS.open("/");
  spiffsFile = spiffsRoot.openNextFile();
  while(spiffsFile)
  {
    tft.print(" ");
    tft.print(spiffsFile.name());
    tft.print(" = ");
    tft.print(spiffsFile.size(), DEC);
    tft.println(" Byte");
    spiffsFile = spiffsRoot.openNextFile();
  }
  spiffsRoot.close(); 
}
